#include <fstream>
#include <iostream>
#include <string>
#include <cstring>
#include <vector>
#include <limits>
#include <iomanip>

typedef std::vector<std::uint8_t> Buf;

bool getBinFile(std::ifstream& in, Buf& out);
size_t binFileSize(std::ifstream& in);
int bindiff(const std::string& f1, const std::string& f2);


int main(int argc, char** argv) {
  if(argc==1) {
    std::clog << "Error: no arguments supplied\n";
    return 1;
  }
  if(strcmp(argv[1],"-h") == 0) {
    std::clog << "Supply names of files to compare as arguments\n";
  }
  if(argc != 3) {
    std::clog << "Error: wrong number of arguments\n";
    return 1;
  }
  return bindiff(argv[1], argv[2]);
}



int bindiff(const std::string& f1, const std::string& f2) {
  Buf dat1;
  Buf dat2;

  std::ifstream in1(f1, std::ios::binary);
  std::ifstream in2(f2, std::ios::binary);

  getBinFile(in1, dat1);
  getBinFile(in2, dat2);

  Buf::iterator p1 = dat1.begin();
  Buf::iterator p2 = dat2.begin();

  std::clog << std::hex << "Comparing " << f1 << " (0x" << dat1.size() << " bytes) & " <<
    f2 << " (0x" << dat2.size() <<" bytes)\n";

  //compare files
  while((p1-dat1.begin()) < dat1.size() && (p2-dat2.begin() < dat2.size()))  {
    if(*p1 == *p2) {
      //match
      p1++;
      p2++;

    }
    else {
      std::clog << "Diff found: ";
      // mismatch:
      // determine which file has extra bytes
      // advance pointers to next matching byte
      // print information to console
      //
      // this assumes that the padding is 2 bytes

      if(*p1 == *(p2+2)) {
        std::clog << std::setfill('0');
        std::clog << "File 2 padded, beginning at 0x" << (p2-dat2.begin()) << " (pad: 0x";
        std::clog << std::setw(2) << +*p2 << std::setw(2) << +*(p2+1) << ")\n";
        p1++;
        p2+=3;
      }
      else if(*(p1+2) == *p2) {
        std::clog << std::setfill('0');
        std::clog << "File 1 padded, beginning at 0x" << (p1-dat1.begin()) << " (pad: 0x";
        std::clog << std::setw(2) << +*p1 << std::setw(2) << +*(p1+1) << ")\n";
        p1+=3;
        p2++;

      }
      else if(*(p1+1) == *(p2+1)) {
        std::clog << "Single modified byte: f1:0x"<<(p1-dat1.begin())<<", f2:0x";
        std::clog << (p2-dat2.begin())<< ". f1:0x"<<+*p1<<",f2:0x"<<+*p2<<"\n";
        p1++;
        p2++;
      }
      else {
        std::clog << "unknown discrepancy: f1:0x"<<(p1-dat1.begin())<<", f2:0x";
        std::clog << (p2-dat2.begin()) << "\n";
        return 2;
      }


    }

  }
  return 0;
}


bool getBinFile(std::ifstream& in, Buf& out) {
	if(!in.is_open())
		return false;
	auto size = binFileSize(in);
	out.resize(size);
	in.read(reinterpret_cast<char*>(out.data()), size);
	return true;
}

size_t binFileSize(std::ifstream& in) {
	if(!in.is_open())
		return false;
	in.ignore( std::numeric_limits<std::streamsize>::max() );
	std::streamsize length = in.gcount();
	in.clear();   //  Since ignore will have set eof.
	in.seekg( 0, std::ios_base::beg );
	return length;
}
